package edu.njit.decaf3;

import edu.njit.decaf3.helper.JSON;
import edu.njit.decaf3.structure.Simulation;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javolution.util.FastMap;
import org.jgrapht.alg.DijkstraShortestPath;
import org.jgrapht.graph.DefaultDirectedWeightedGraph;
import org.jgrapht.graph.DefaultWeightedEdge;


/**
 * DECAF - Dependability Evaluator of Cascading Failures Revision 3
 *
 * @author Sashank Tadepalli
 */
public class ThresholdAnalysis {
	public static final short CSV = 1;
	public static final short PRETTY = 2;

	private String CSV_OUTPUT = "";
	private short printType = 2;

	private Simulation sim;
	private DefaultDirectedWeightedGraph<Integer, DefaultWeightedEdge> directedGraph;
	private double[][] treeRateCache;
	private List<DefaultWeightedEdge>[][] pathCache;
	private FastMap<DefaultWeightedEdge, Double> edgeCache;
	private final String DELIM = ",";
	private final String NL_DELIM = "\n";

	private String optimalRootType;
	private int optimalK;
	private double optimalRateThreshold;
	private double optimalProbability;
	public double AnalyzeTime;

	/**
	 * @throws java.io.FileNotFoundException
	 * @throws java.io.IOException
	 */
	public ThresholdAnalysis() throws Exception {
		this(new File("data/simulation.model.json"));
	}

	/**
	 * @param inputFileLocation
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public ThresholdAnalysis(String inputFileLocation) throws Exception {
		this(new File(inputFileLocation));
	}

	/**
	 * @param inputFile
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public ThresholdAnalysis(File inputFile) throws Exception {
		// Assert inputFile exists
		if (!inputFile.exists()) {
			throw new FileNotFoundException(inputFile.getPath() + " cannot be found in " + new File(".").getCanonicalPath());
		}
		// Load Json Data
		sim = JSON.parse(inputFile);
		init();
	}

	/**
	 * @param s
	 */
	public ThresholdAnalysis(Simulation s) {
		sim = s;
		init();
	}

	private void init() {
		// Build model
		directedGraph = new DefaultDirectedWeightedGraph<>(DefaultWeightedEdge.class);
		edgeCache = new FastMap<>();
		for (int i = 0; i < sim.components.length; i++)
			directedGraph.addVertex(i);

		// Initialize treeRateCache
		treeRateCache = new double[sim.cascading.length][];
		pathCache = new List[sim.cascading.length][];
		for (int i = 0; i < sim.cascading.length; i++) {
			treeRateCache[i] = new double[sim.components[i].redundancy];
			pathCache[i] = new List[sim.components[i].redundancy];

			for (int j = 0; j < sim.cascading[i].length; j++) {
				DefaultWeightedEdge t = directedGraph.addEdge(i, j);
				double w = Math.abs(Math.log10(1D - sim.cascading[i][j]));
				w = w == 0 ? Double.POSITIVE_INFINITY : w;
				directedGraph.setEdgeWeight(t, w);
				edgeCache.put(t, sim.cascading[i][j]);
			}
		}
	}

	public double getOptimalRateThreshold() {
		long t = System.nanoTime();
		CSV_OUTPUT += "Root" + DELIM + "k" + DELIM + "Tree" +
				DELIM + "Rate" + DELIM + "Probability" + DELIM + "MTTF" +
				DELIM + "U Trees" + DELIM + "F Trees" + NL_DELIM;

		// Calculate probabilities per component
		for (int i = 0; i < sim.cascading.length; i++) {
			int numRequired = sim.components[i].redundancy - sim.components[i].required + 1;
			int[] n = new int[sim.components.length];

			// Start with p_k where k = 0 until k = numRequired
			for (int k = 1; k < numRequired + 1; k++) {

				// Initialize fresh n[] and sizesInTransition[] for each comp iteration
				for (int j = 0; j < n.length; j++)
					n[j] = sim.components[j].redundancy;

				CSV_OUTPUT += "" + sim.components[i].type + DELIM + k + DELIM;

				if (calculateProbability(n, getTransitionSizes(k, numRequired), i)) {
					optimalK = k;
				}
			}
		}
		AnalyzeTime = (System.nanoTime() - t) / 1e9f;
		return optimalRateThreshold;
	}

	private boolean calculateProbability(int[] n, int[] sizesInTransition, int rootIdx) {
		double totalProb = 1;
		double totalRate = Double.POSITIVE_INFINITY;

		// Iterate through each tree size, grab rate and multiply into totalProb
		// Also, keep track of smallest tree rate
		for (int i = 0; i < sizesInTransition.length; i++) {
			double tmpRate = getRate(sizesInTransition[i], rootIdx);
			String treePath = pathCache[rootIdx][sizesInTransition[i] - 1] + "";
			totalProb *= n[rootIdx];

			// Adjust n[] and push 1 - phi
			n[rootIdx]--;
			adjustN(n, treePath);

			totalProb *= tmpRate / (getDenomA(n, sim) + (i == 0 ? 0 : getDenomB(n, sim)));
			//System.out.println("\n" + tmpRate + "\n" + getDenomA(n, sim) + "," + totalProb + "\n");

			// Find smallest tree rate
			totalRate = Math.min(tmpRate, totalRate);

			if (pathCache[rootIdx][sizesInTransition[i] - 1] == null) {
				CSV_OUTPUT += "N/A";
			} else {
				String resultString = treePath
						.replaceAll("(\\(|\\[|\\]|\\)|\\s)", "")
						.replaceAll(",", "->")
						.replaceAll(":", "->").replaceAll("->,", ";");
				resultString = resultString.length() > 0 ? "->" + resultString : resultString;

				for (int j = 0; j < sim.components.length; j++)
					resultString = resultString.replaceAll(j + "", sim.components[j].type);

				CSV_OUTPUT += sim.components[rootIdx].type + resultString + ";";
			}
		}

		// Ensure n[] is a valid state
		for (int l : n)
			if (l < 0)
				totalProb = 0;

		if (n[rootIdx] != 0)
			totalProb = 0;

		CSV_OUTPUT += DELIM + totalRate + DELIM;
		CSV_OUTPUT += totalProb + DELIM + NL_DELIM;

		if (totalProb > optimalProbability) {
			optimalProbability = totalProb;
			optimalRateThreshold = totalRate;
			optimalRootType = sim.components[rootIdx].type;
			return true;
		}
		return false;
	}

	/**
	 * @param t
	 * @param path
	 */
	private void adjustN(int[] t, String path) {
		for (int j = 0; j < path.length(); j++) {
			try {
				int currIdx = Integer.parseInt(String.valueOf(path.charAt(j)));
				t[currIdx]--;
			} catch (Exception e) {
			}
		}
	}

	/**
	 * @param n
	 * @param sim
	 * @return
	 */
	private double getDenomA(int[] n, Simulation sim) {
		double result = 0;
		for (int i = 0; i < sim.components.length; i++) {
			result += n[i] * sim.components[i].failure[0];
		}
		return result;
	}

	/**
	 * @param n
	 * @param sim
	 * @return
	 */
	private double getDenomB(int[] n, Simulation sim) {
		double a = 0, b = 0;
		for (int i = 0; i < sim.components.length; i++) {
			a += n[i] * sim.components[i].repair[0];
			b += n[i];
		}
		if (b == 0)
			return 0;
		return a / b;
	}

	/**
	 * @param k
	 * @param r
	 * @return
	 */
	private int[] getTransitionSizes(int k, int r) {
		int sLow = (int) Math.floor((double) r / k);
		int sHigh = (int) Math.ceil((double) r / k);
		int tmp;
		int[] transitionSizes = new int[k];

		Arrays.fill(transitionSizes, sHigh);

		if (sLow == sHigh) {
			return transitionSizes;
		}
		tmp = sHigh * k;
		for (int z = 0; z < k; z++) {
			if (tmp != r) {
				transitionSizes[z]--;
				tmp--;
			}
		}
		return transitionSizes;
	}

	/**
	 * @param failedComps
	 * @param rootIndex
	 * @return
	 */
	private double getRate(int failedComps, int rootIndex) {
		failedComps--;
		if (treeRateCache[rootIndex][failedComps] != 0) {
			return treeRateCache[rootIndex][failedComps];
		}
		if (failedComps == 0) {
			pathCache[rootIndex][failedComps] = new ArrayList<>();
			return treeRateCache[rootIndex][failedComps] = sim.components[rootIndex].failure[0];
		}

		double largestTreeRate = 0;
		List<DefaultWeightedEdge> bestPath = null;
		List<DefaultWeightedEdge> finalPath = new ArrayList<>();
		for (int i = 0; i < sim.components.length; i++) {
			List<DefaultWeightedEdge> path = DijkstraShortestPath.findPathBetween(directedGraph, i, rootIndex);

			// Path not possible?
			if (path.size() == 0) continue;

			double tmpRate = sim.cascading[rootIndex][i];
			for (DefaultWeightedEdge e : path) {
				tmpRate *= edgeCache.get(e);
			}
			if (tmpRate > largestTreeRate) {
				largestTreeRate = tmpRate;
				bestPath = path;
			}
		}
		if (bestPath == null) {
			pathCache[rootIndex][failedComps] = new ArrayList<>();
			return treeRateCache[rootIndex][failedComps] = sim.components[rootIndex].failure[0];
		}
		for (int i = failedComps; i > 0; i--) {
			finalPath.addAll(bestPath);
		}
		pathCache[rootIndex][failedComps] = finalPath;
		return treeRateCache[rootIndex][failedComps] =
				sim.components[rootIndex].failure[0] * Math.pow(largestTreeRate, failedComps);
	}

	/**
	 * @return
	 */
	public short getPrintType() {
		return printType;
	}

	/**
	 * @param printType
	 */
	public void setPrintType(short printType) {
		this.printType = printType;
	}

	@Override
	public String toString() {
		switch (printType) {
			case CSV:
				return CSV_OUTPUT;
			case PRETTY:
				if (optimalRootType.isEmpty())
					return "Please run getOptimalRateThreshold method first";
				return "Highest probability at root type " + optimalRootType + ", requiring " +
						optimalK + " failure transition(s):\n" + optimalRateThreshold;
			default:
				return "";
		}
	}
}